void	*xmalloc(size_t, const char *);
void	*xcalloc(size_t, size_t, const char *);
char	*xstrdup(const char *, const char *);
void	*xmemdup(const char *, size_t, const char *);
char	*strip(char *);
int	 lowercase(char *, const char *, size_t);
